from pact_test.either import *
from pact_test.matchers.matcher import *


def match(interaction, pact_response):
    return _match_status(interaction, pact_response)\
        .concat(_match_headers, pact_response)\
        .concat(_match_body, pact_response)


def _match_status(interaction, pact_response):
    expected = interaction['response'].get('status')
    actual = pact_response.status

    if expected is None or actual == expected:
        return Right(interaction)
    return Left(build_error_message('status', expected, actual))


def _match_headers(interaction, pact_response):
    expected = interaction['response'].get('headers', {})
    actual = _to_dict(pact_response.headers)

    insensitive_expected = {k.upper(): v for (k, v) in expected.items()}
    insensitive_actual = {k.upper(): v for (k, v) in actual.items()}

    if is_subset(insensitive_expected, insensitive_actual):
        return Right(interaction)
    return Left(build_error_message('headers', expected, actual))


def _match_body(interaction, pact_response):
    expected = interaction['response'].get('body')
    actual = pact_response.body

    if expected is None and actual is None:
        return Right(interaction)

    if is_string(expected) and is_string(actual) and expected == actual:
        return Right(interaction)

    if type(expected) is dict and type(actual) is dict and match_dicts_all_keys_and_values(expected, actual):
            return Right(interaction)

    return Left(build_error_message('body', expected, actual))


def _to_dict(headers):
    d = {}
    for h in headers:
        d[h[0]] = h[1]
    return d



import requests
from pact_test.either import *
from pact_test.constants import *
from pact_test.utils.logger import *
from pact_test.models.response import PactResponse


def execute_interaction_request(url, port, interaction):
    url = _build_url(url, port, interaction)
    method = interaction[REQUEST].get('method', 'GET')
    server_response = _server_response(method, url=url)

    if type(server_response) is Right:
        headers = _parse_headers(server_response.value)
        content_type = _get_content_type(headers)
        out = Right(PactResponse(
            status=server_response.value.status_code,
            headers=headers,
            body=_parse_body(server_response.value, content_type)
        ))
        return out

    return server_response


def _server_response(method, url):
    try:
        return Right(requests.request(method, url=url))
    except Exception as e:
        return Left(str(e))


def _parse_body(server_response, content_type):
    return server_response.json() if JSON in content_type else server_response.text


def _parse_headers(server_response):
    headers = []
    server_headers = server_response.headers or []
    for key in server_headers:
        headers.append((key, server_headers[key]))
    return headers


def _build_url(url, port, interaction):
    path = interaction[REQUEST].get('path', '')
    query = interaction[REQUEST].get('query', '')

    test_port = str(port) if port else str(80)
    return 'http://' + url + ':' + test_port + path + query


def _get_content_type(headers):
    content_type = TEXT
    types = list(filter(lambda h: h[0].upper() == 'CONTENT-TYPE', headers))
    if types:
        content_type = types[0][1]
    return content_type



from pact_test.either import *
from pact_test.constants import *
from pact_test.utils.logger import error
from pact_test.matchers.response_matcher import match
from pact_test.clients.http_client import execute_interaction_request


def verify_state(interaction, pact_helper, test_instance):
    state = find_state(interaction, interaction['description'], test_instance)
    if type(state) is Right:
        state.value()
        output = _execute_request(pact_helper, interaction)
        if type(output) is Right:
            response_verification = match(interaction, output.value)
            output = _build_state_response(state, interaction['description'], response_verification)
            return output
        error(output.value)
        return output
    error(state.value)
    return state


def _execute_request(pact_helper, interaction):
    url = pact_helper.test_url
    port = pact_helper.test_port
    return execute_interaction_request(url, port, interaction)


def _build_state_response(state, description, response_verification):
    if type(response_verification) is Right:
        return Right(_format_message(state.value.state, description, PASSED, []))
    else:
        errors = [response_verification.value]
        return Left(_format_message(state.value.state, description, FAILED, errors))


def find_state(interaction, description, test_instance):
    state = interaction[PROVIDER_STATE]
    for s in test_instance.states:
        if s.state == state:
            return Right(s)
    message = 'Missing state implementation for "' + state + '"'
    return Left(_format_message(state, description, FAILED, [message]))


def _format_message(state, description, status, errors):
    return {'state': state, 'description': description, 'status': status, 'errors': errors}



from pact_test.models.request import PactRequest
from pact_test.models.response import PactResponse


REQUEST = 'request'
RESPONSE = 'response'


def build_request_from_interaction(interaction):
    request = interaction[REQUEST]

    return PactRequest(
        body=request.get('body', None),
        path=request.get('path', None),
        query=request.get('query', None),
        method=request.get('method', None),
        headers=_build_headers_from_pact(interaction, REQUEST)
    )


def build_response_from_interaction(interaction):
    response = interaction[RESPONSE]

    return PactResponse(
        body=response.get('body', None),
        status=response.get('status', None),
        headers=_build_headers_from_pact(interaction, RESPONSE)
    )


def _build_headers_from_pact(pact, request_or_response):
    headers = []
    pact_headers = pact[request_or_response].get('headers', None)
    for h in (h for h in (pact_headers if pact_headers else [])):
        headers.append((h, pact_headers[h]))
    return headers



import json
from pact_test.either import *
from pact_test.utils.logger import *
from pact_test.models.request import PactRequest
from pact_test.models.response import PactResponse
from pact_test.servers.mock_server import MockServer
from pact_test.matchers.request_matcher import match


def verify_request(decorated_method, port=9999):
    mock_response = build_expected_response(decorated_method)
    expected_request = build_expected_request(decorated_method)

    mock_server = MockServer(mock_response=mock_response, port=port)
    mock_server.start()
    decorated_method()
    mock_server.shutdown()
    report = mock_server.report()

    if len(report) is 0:
        out = {
            'status': 'FAILED',
            'providerState': decorated_method.given,
            'description': decorated_method.upon_receiving,
            'message': 'Missing request(s) for "' + format_message(decorated_method) + '"',
            'expected': expected_request.__dict__,
            'actual': None
        }
        return Left(out)

    actual_request = build_actual_request(report[0])
    matching_result = match(actual_request, expected_request)

    if type(matching_result) is Right:
        resp = mock_response.__dict__
        resp['body'] = json.loads(mock_response.__dict__['body'])
        out = {
            'status': 'PASSED',
            'providerState': decorated_method.given,
            'description': decorated_method.upon_receiving,
            'request': actual_request.__dict__,
            'response': resp
        }
        return Right(out)
    else:
        reason = matching_result.value.copy()
        reason.update({
            'providerState': decorated_method.given,
            'description': decorated_method.upon_receiving
        })
        return Left(reason)


def build_expected_response(decorated_method):
    return PactResponse(
        body=decorated_method.will_respond_with.get('body'),
        status=decorated_method.will_respond_with.get('status'),
        headers=decorated_method.will_respond_with.get('headers')
    )


def build_expected_request(decorated_method):
    return PactRequest(
        method=decorated_method.with_request.get('method'),
        body=decorated_method.with_request.get('body'),
        headers=decorated_method.with_request.get('headers'),
        path=decorated_method.with_request.get('path'),
        query=decorated_method.with_request.get('query')
    )


def build_actual_request(request):
    return PactRequest(
        path=request.get('path'),
        query=request.get('query'),
        method=request.get('method'),
        body=request.get('body'),
        headers=request.get('headers')
    )


def format_message(decorated_method):
    return 'given ' + \
           decorated_method.given + \
           ', upon receiving ' + \
           decorated_method.upon_receiving



from setuptools import setup
from setuptools import find_packages

setup(
    name='pact-test',
    version='0.3.70',
    author='Guido Barbaglia',
    author_email='guido.barbaglia@gmail.com',
    packages=find_packages(),
    license='LICENSE',
    long_description=open('README.rst').read(),
    description='Python implementation for Pact (http://pact.io/)',
    install_requires=['requests'],
    setup_requires=['pytest-runner'],
    tests_require=[
        'pytest>=3.0',
        'pytest-pep8',
        'pytest-sugar',
        'pytest-mock'
    ],
    url='https://github.com/Kalimaha/pact-test/',
    scripts=['bin/pact-test'],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Testing'
    ]
)
